#define QT_FEATURE_wayland_client 1
#define QT_FEATURE_drm_egl_server 1
#define QT_FEATURE_libhybris_egl_server -1
#define QT_FEATURE_wayland_brcm -1
#define QT_FEATURE_wayland_datadevice 1
#define QT_FEATURE_wayland_egl 1
#define QT_FEATURE_xcomposite_egl 1
#define QT_FEATURE_xcomposite_glx -1
