#ifndef QT5DWAYLANDPLUGIN_H
#define QT5DWAYLANDPLUGIN_H

#include "qt5dwayland-plugin_global.h"

class QT5DWAYLANDPLUGINSHARED_EXPORT Qt5dwaylandplugin
{

public:
    Qt5dwaylandplugin();
};

#endif // QT5DWAYLANDPLUGIN_H
