#ifndef QT5DWAYLANDPLUGIN_GLOBAL_H
#define QT5DWAYLANDPLUGIN_GLOBAL_H

#include <QtCore/qglobal.h>

#if defined(QT5DWAYLANDPLUGIN_LIBRARY)
#  define QT5DWAYLANDPLUGINSHARED_EXPORT Q_DECL_EXPORT
#else
#  define QT5DWAYLANDPLUGINSHARED_EXPORT Q_DECL_IMPORT
#endif

#endif // QT5DWAYLANDPLUGIN_GLOBAL_H
